@@ -1,24 +1,30 @@
+'''
+    Defines the ArchiveAPI and relevant search/order entities.
+'''
+
 from __future__ import annotations
+import enum
 import json
-import string
-import textwrap
 import typing
 import requests
 
-from dataclasses import dataclass
 from datetime import date, datetime
-from .common import ArlulaObject
+
+from .order import Order
+from .common import ArlulaObject, Band, Bundle, License, SortDefinition, get_bundle_key, get_license_href
 from .auth import Session
 from .exception import ArlulaAPIException
-from .orders import DetailedOrderResult
 from .util import parse_rfc3339, calculate_price, remove_none, simple_indent
 
-Polygon = typing.Union[typing.List[typing.List[typing.List[float]]], str]
+Polygon = typing.List[typing.List[typing.List[float]]]
 
 class CenterPoint(ArlulaObject):
     data: dict
     long: float
+    """Center longitude"""
+
     lat: float
+    """Center latitude"""
 
     def __init__(self, data):
         self.data = data
@@ -34,7 +40,10 @@ class CenterPoint(ArlulaObject):
 class Percent(ArlulaObject):
     data: dict
     scene: float
+    """percent of the whole scene the polygon represents"""
+
     search: float
+    """percent of the original search area (if bounding box) the AOI represents. This will be -1 for a point search, and may be greater than 100% if the original search is less than the suppliers minimum order"""
 
     def __init__(self, data):
         self.data = data
@@ -50,8 +59,13 @@ class Percent(ArlulaObject):
 class Overlap(ArlulaObject):
     data: dict
     area: float
+    """area in sq km of the overlap between search and result"""
+
     percent: Percent
+    """details percentage coverage"""
+
     polygon: typing.List[typing.List[typing.List[float]]]
+    """polygon of overlap area. This polygon is what will be ordered if the supplier supports ordering an area of interest"""
 
     def __init__(self, data):
         self.data = data
@@ -69,109 +83,61 @@ class Overlap(ArlulaObject):
         f"{str(self.percent)}\n"\
         f"Geometry: {self.polygon}", 0, 2)
 
-class License(ArlulaObject):
-    data: dict
-    name: str
-    href: str
-    loading_percent: float
-    loading_amount: int
-
-    def __init__(self, data):
-        self.data = data
-        self.name = data["name"]
-        self.href = data["href"]
-        self.loading_percent = data["loadingPercent"]
-        self.loading_amount = data["loadingAmount"]
-
-    def dict(self) -> dict:
-        return self.data
-
-    def __str__(self) -> str:
-        return simple_indent(
-        f"License ({self.href}):\n"\
-        f"Name: {self.name}\n"\
-        f"Loading: {self.loading_amount} US Cents + {self.loading_percent}%\n", 0, 2)
-
-class Band(ArlulaObject):
-    data: dict
-    name: str
-    id: str
-    min: float
-    max: float
-
-    def __init__(self, data):
-        self.data = data
-        self.name = data["name"]
-        self.id = data["id"]
-        self.min = data["min"]
-        self.max = data["max"]
-
-    def centre(self) -> float:
-        '''
-            Get the band centre wavelength
-        '''
-        return (self.max - self.min)/2
-
-    def width(self) -> float:
-        '''
-            Get the band width
-        '''
-        return self.max - self.min
-
-    def dict(self) -> dict:
-        return self.data
-
-    def __str__(self) -> str:
-        return simple_indent(
-            f"Band ({self.id}):\n"\
-            f"Name: {self.name}\n"\
-            f"Bandwidth: {self.min}nm - {self.max}nm\n", 0, 2)
-
-class Bundle(ArlulaObject):
-    data: dict
-    name: str
-    key: str
-    bands: typing.List[str]
-    price: int
-
-    def __init__(self, data):
-        self.data = data
-        self.name = data["name"]
-        self.key = data["key"]
-        self.bands = data["bands"]
-        self.price = data["price"]
-
-    def dict(self) -> dict:
-        return self.data
-
-    def __str__(self) -> str:
-        bands = 'all' if len(self.bands) == 0 else '\n'.join(self.bands)
-        return simple_indent(
-            f"Bundle ({self.key}):\n"\
-            f"Name: {self.name}\n"\
-            f"Bands: {bands}\n"\
-            f"Price: {self.price} US Cents\n", 0, 2)
-
 class SearchResult(ArlulaObject):
     data: dict
     scene_id: str
+    """ID for the suppliers image capture, can be used to identify the same source imagery between searches"""
+
     supplier: str
+    """identifies the supplier of the imagery, used in ordering the imagery"""
+
     platform: str
+    """The platform which captured the imagery, generally identifies the constellation, or specific satellite that captured the imagery."""
+
     date: datetime
+    """date the imagery was captured"""
+
     thumbnail: str
+    """URL of a low resolution JPEG thumbnail of the imagery that will be provided"""
+
     cloud: float
+    """estimated percentage of the imagery that is cloud cover"""
+
     off_nadir: float
+    """The degrees away from Nadir (straight down) the satellite was oriented during capture"""
+
     gsd: float
+    """spatial resolution of the imagery in meters per pixel"""
+
     bands: typing.List[Band]
+    """List of the Spectral Bands captured in this scene"""
+
     area: float
+    """area the scene covers in square kilometers"""
+
     center: CenterPoint
+    """center coordinates of the imagery"""
+
     bounding: typing.List[typing.List[typing.List[float]]]
+    """bounding polygon of the imagery in the geoJSON ([long, lat]) notation"""
+
     overlap: Overlap
+    """overlap between the imagery and the interest area, or an area constructed from it to meet minimum order requirements"""
+
     fulfillment_time: float
+    """estimated time to fulfill an order of this imagery in hours, 0 is instant"""
+
     ordering_id: str
+    """ordering ID for this scene and Area Of Interest, used to order the imagery"""
+
     bundles: typing.List[Bundle]
-    license: typing.List[License]
+    """ordering bundles representing the available ways to order the imagery"""
+
+    licenses: typing.List[License]
+    """License options this imagery may be purchased under, and the terms and pricing that apply"""
+
     annotations: typing.List[str]
+    """annotates result with information, such as what modifications were made to your search to make a valid order"""
 
     def __init__(self, data):
         self.data = data
@@ -198,14 +164,14 @@ class SearchResult(ArlulaObject):
             
         self.bundles = []
         self.bundles += [Bundle(b) for b in data["bundles"]]
-        self.license = []
-        self.license += [License(l) for l in data["license"]]
+        self.licenses = []
+        self.licenses += [License(l) for l in data["licenses"]]
 
         self.annotations = []
         if "annotations" in data:
             self.annotations = data["annotations"]
 
-    def calculate_price(self, license_href: string, bundle_key: string) -> int:
+    def calculate_price(self, license_href: str, bundle_key: str) -> int:
         '''
             Wrapper for util.calculate_price, returns price in US Cents. Raises error in the case of invalid license_name or bundle_key
         '''
@@ -214,11 +180,11 @@ class SearchResult(ArlulaObject):
         license = None
         
         for b in self.bundles:
-            if bundle.key == bundle_key:
+            if b.key == bundle_key:
                 bundle = b
         
-        for l in self.license:
-            if license.href == license_href:
+        for l in self.licenses:
+            if l.href == license_href:
                 license = l
 
         if bundle == None:
@@ -236,7 +202,7 @@ class SearchResult(ArlulaObject):
 
         bundles = simple_indent(''.join([str(b) for b in self.bundles]), 2, 2)
         bands = simple_indent(''.join([str(b) for b in self.bands]), 2, 2)
-        license = simple_indent(''.join([str(l) for l in self.license]), 2, 2)
+        licenses = simple_indent(''.join([str(l) for l in self.licenses]), 2, 2)
         return simple_indent(
             f"Result ({self.ordering_id}):\n"\
             f"Scene ID: {self.scene_id}\n"\
@@ -256,16 +222,17 @@ class SearchResult(ArlulaObject):
             f"{bands}"
             f"Bundles: \n"\
             f"{bundles}"
-            f"License: \n"\
-            f"{license}"
+            f"Licenses: \n"\
+            f"{licenses}"
             f"Annotations: {', '.join(self.annotations)}\n", 0, 2)
 
 class SearchResponse(ArlulaObject):
     data: dict
-    state: string
+    state: str
     errors: typing.List[str]
     warnings: typing.List[str]
     results: typing.List[SearchResult]
+    """Results for the search conducted."""
 
     def __init__(self, data):
         self.data = data
@@ -293,20 +260,65 @@ class SearchResponse(ArlulaObject):
     def dict(self) -> dict:
         return self.data
 
+class ArchiveSearchSortFields(str, enum.Enum):
+    """
+        An enumeration of fields that can be sorted by on archive search requests.
+    """
+
+    scene_id = "sceneID"
+    supplier = "supplier"
+    date = "date"
+    cloud = "cloud"
+    off_nadir = "offNadir"
+    gsd = "gsd"
+    area = "area"
+    overlap = "overlap"
+    overlap_area = "overlap.area"
+    overlap_percent = "overlap.percent"
+    fulfillment = "fulfillment"
+
 class SearchRequest():
     start: date
+    """Date of interest, or start of an interest period"""
+
     gsd: float
+    """Desired ground sample distance of imagery, all imagery will be of this GSD or better"""
+    
     end: date
+    """End of an interest period"""
+
     lat: float
+    """Latitude of the point of interest for which imagery is requested"""
+    
     long: float
+    """Longitude of the point of interest for which imagery is requested"""
+
     north: float
+    """Northern boundary of an interest area for which imagery is requested"""
+    
     south: float
+    """Southern boundary of an interest area for which imagery is requested"""
+    
     east: float
+    """Eastern boundary of an interest area for which imagery is requested"""
+    
     west: float
+    """Western boundary of an interest area for which imagery is requested"""
+    
     polygon: Polygon
+    """Polygon demarking an interest area for which imagery is requested."""
+    
     supplier: str
+    """Restrict results to only the given (space separated) suppliers"""
+    
     off_nadir: float
+    """Only return imagery with the off nadir angle during capture less than the specified criteria"""
+    
     cloud: float
+    """Only return imagery with average total cloud percentage less than the specified criteria"""
+    
+    sort_definition: SortDefinition[ArchiveSearchSortFields]
+    """The desired field to sort results by, and if that sort should be ascending or descending"""
 
     def __init__(self, start: date,
             gsd: float,
@@ -320,7 +332,9 @@ class SearchRequest():
             west: typing.Optional[float] = None,
             supplier: typing.Optional[str] = None,
             off_nadir: typing.Optional[float] = None,
-            polygon: Polygon = None):
+            polygon: typing.Optional[Polygon] = None,
+            sort_definition: typing.Optional[SortDefinition[ArchiveSearchSortFields]] = None,
+        ):
         self.start = start
         self.cloud = cloud
         self.gsd = gsd
@@ -334,6 +348,7 @@ class SearchRequest():
         self.supplier = supplier
         self.off_nadir = off_nadir
         self.polygon = polygon
+        self.sort_definition = sort_definition
 
     def set_point_of_interest(self, lat: float, long: float) -> "SearchRequest":
         self.lat = lat
@@ -352,6 +367,7 @@ class SearchRequest():
         return self
 
     def set_supplier(self, supplier: str) -> "SearchRequest":
+        """Filter the suppliers to search. To set multiple, use a space separated list."""
         self.supplier = supplier
         return self
 
@@ -379,6 +395,10 @@ class SearchRequest():
     def set_maximum_cloud_cover(self, cloud: float) -> "SearchRequest":
         self.cloud = cloud
         return self
+    
+    def set_sort_definition(self, sort_definition: SortDefinition[ArchiveSearchSortFields]):
+        self.sort_definition = sort_definition
+        return self
 
     def valid_point_of_interest(self) -> bool:
         return self.lat != None and self.long != None
@@ -390,61 +410,108 @@ class SearchRequest():
         return (self.valid_area_of_interest() or self.valid_point_of_interest) and self.start != None and self.gsd != None
     
     def dict(self):
-        param_dict = {
+        d = {
             "start": str(self.start) if self.start != None else None, 
             "end": str(self.end) if self.end != None else None,
-            "gsd": self.gsd, "cloud": self.cloud,
-            "lat": self.lat, "long": self.long,
-            "north": self.north, "south": self.south, "east": self.east, 
-            "west": self.west, "supplier": self.supplier, "off-nadir": self.off_nadir,
-            "polygon": json.dumps(self.polygon) if isinstance(self.polygon, list) else self.polygon}
+            "gsd": self.gsd, 
+            "cloud": self.cloud,
+            "offNadir": self.off_nadir,
+            "supplier": self.supplier,
+        }
 
-        query_params = {k: v for k, v in param_dict.items()
-            if v is not None}
+        # Add the polygon if not None
+        if self.polygon != None:
+            d["polygon"] = self.polygon if isinstance(self.polygon, list) else self.polygon
+        # Add boundingBox if all related not None
+        elif self.north != None and self.east != None and self.west != None and self.south != None:
+            d["boundingBox"] = {
+                "north": self.north,
+                "east": self.east,
+                "west": self.west,
+                "south": self.south,
+            }
+        # Add latLong if all related not None
+        elif self.lat != None and self.long != None:
+            d["latLong"] = {
+                "latitude": self.lat,
+                "longitude": self.long,
+            }
+        
+        if self.sort_definition is not None:
+            d["sort"] = self.sort_definition.dict()
 
-        return remove_none(query_params)
+        return remove_none(d)
 
-class OrderRequest(ArlulaObject):
+class ArchiveOrderRequest(ArlulaObject):
 
     id: str
+    """Unique ID of the imagery to purchase, provided in the search endpoint"""
+
     eula: str
+    """The eula string provided in the search results to confirm acceptance. (the href of the license of interest)"""
+
     bundle_key: str
+    """The key of the bundle you wish to purchase from the scene"""
+
     webhooks: typing.List[str]
+    """A list of http/https addresses that the order details will be sent to once the order is complete and ready for collection"""
+
     emails: typing.List[str]
+    """A list of email addresses (strings) that the order details will be sent to once the order is complete and ready for collection"""
+
+    team: str
+    """The ID of the team to attach the order to, if other than the default team for the API (used to control shared access to the order)"""
+
+    payment: str
+    """The ID of the payment method to charge this order to (if not free). If not specified, the APIs default billing method will be charged."""
 
     def __init__(self,
             id: str,
-            eula: str,
-            bundle_key: str,
+            license: typing.Union[str, License],
+            bundle: typing.Union[str, Bundle],
             webhooks: typing.Optional[typing.List[str]] = [],
             emails: typing.Optional[typing.List[str]] = [],
-            team: typing.Optional[str] = None):
+            team: typing.Optional[str] = None,
+            payment: typing.Optional[str] = None):
         self.id = id
-        self.eula = eula
-        self.bundle_key = bundle_key
+        self.eula = get_license_href(license)
+        self.bundle_key = get_bundle_key(bundle)
         self.webhooks = webhooks
         self.emails = emails
         self.team = team
+        self.payment = payment
+
+    def set_bundle(self, bundle: typing.Union[str, Bundle]) -> ArchiveOrderRequest:
+        self.bundle_key = get_bundle_key(bundle)
+        return self
     
-    def add_webhook(self, webhook: str) -> "OrderRequest":
+    def set_eula(self, license: typing.Union[str, License]) -> ArchiveOrderRequest:
+        self.eula = get_license_href(license)
+        return self
+    
+    def add_webhook(self, webhook: str) -> "ArchiveOrderRequest":
         self.webhooks.append(webhook)
         return self
     
-    def set_webhooks(self, webhooks: typing.List[str]) -> "OrderRequest":
+    def set_webhooks(self, webhooks: typing.List[str]) -> "ArchiveOrderRequest":
         self.webhooks = webhooks
         return self
 
-    def add_email(self, email: str) -> "OrderRequest":
+    def add_email(self, email: str) -> "ArchiveOrderRequest":
         self.emails.append(email)
         return self
     
-    def set_emails(self, emails: typing.List[str]) -> "OrderRequest":
+    def set_emails(self, emails: typing.List[str]) -> "ArchiveOrderRequest":
         self.emails = emails
         return self
 
-    def set_team(self, team: str) -> "OrderRequest":
+    def set_team(self, team: str) -> "ArchiveOrderRequest":
         self.team = team
         return self
+    
+    def set_payment(self, payment: str) -> "ArchiveOrderRequest":
+        self.payment = payment
+        return payment
 
     def valid(self) -> bool:
         return self.id != None and self.eula != None and self.bundle_key != None
@@ -457,8 +524,84 @@ class OrderRequest(ArlulaObject):
             "webhooks": self.webhooks,
             "emails": self.emails,
             "team": None if self.team == "" else self.team,
+            "payment": None if self.payment == "" else self.payment,
         })
 
+class ArchiveBatchOrderRequest():
+
+    orders: typing.List[ArchiveOrderRequest]
+    """Orders to be placed in batch request."""
+
+    webhooks: typing.List[str]
+    """A list of http/https addresses that the order details will be sent to once the order is complete and ready for collection"""
+
+    emails: typing.List[str]
+    """A list of email addresses (strings) that the order details will be sent to once the order is complete and ready for collection"""
+
+    team: str
+    """The ID of the team to attach the order to, if other than the default team for the API (used to control shared access to the order)"""
+
+    payment: str
+    """The ID of the payment method to charge this order to (if not free). If not specified, the APIs default billing method will be charged."""
+
+    def __init__(
+        self, 
+        orders: typing.Optional[typing.List[ArchiveOrderRequest]] = [],
+        webhooks: typing.Optional[typing.List[str]] = [],
+        emails: typing.Optional[typing.List[str]] = [],
+        team: typing.Optional[str] = None,
+        payment: typing.Optional[str] = None):
+
+        self.orders = orders
+        self.webhooks = webhooks
+        self.emails = emails
+        self.team = team
+        self.payment = payment
+
+    def add_order(self, order: ArchiveOrderRequest) -> "ArchiveBatchOrderRequest":
+        self.orders.append(order)
+        return self
+    
+    def set_orders(self, orders: typing.List[ArchiveOrderRequest]) -> "ArchiveBatchOrderRequest":
+        self.orders = orders
+        return self
+
+    def add_webhook(self, webhook: str) -> "ArchiveBatchOrderRequest":
+        self.webhooks.append(webhook)
+        return self
+    
+    def set_webhooks(self, webhooks: typing.List[str]) -> "ArchiveBatchOrderRequest":
+        self.webhooks = webhooks
+        return self
+
+    def add_email(self, email: str) -> "ArchiveBatchOrderRequest":
+        self.emails.append(email)
+        return self
+    
+    def set_emails(self, emails: typing.List[str]) -> "ArchiveBatchOrderRequest":
+        self.emails = emails
+        return self
+
+    def set_team(self, team: str) -> "ArchiveBatchOrderRequest":
+        self.team = team
+        return self
+    
+    def set_payment(self, payment: str) -> "ArchiveBatchOrderRequest":
+        self.payment = payment
+        return self
+
+    def dict(self):
+        d = {
+            "orders": [o.dict() for o in self.orders],
+            "webhooks": self.webhooks,
+            "emails": self.emails,
+            "team": None if self.team == "" else self.team,
+            "payment": None if self.payment == "" else self.payment,
+        }
+
+        return remove_none(d)
+
+
 
 class ArchiveAPI:
     '''
@@ -480,9 +623,9 @@ class ArchiveAPI:
 
         # Send request and handle responses
         response = requests.request(
-            "GET", url,
+            "POST", url,
             headers=self.session.header,
-            params=request.dict())
+            data=json.dumps(request.dict()))
         if response.status_code != 200:
             raise ArlulaAPIException(response)
         else:
@@ -490,7 +633,7 @@ class ArchiveAPI:
             # Construct an instance of `SearchResponse`
             return SearchResponse(resp_data)
 
-    def order(self, request: OrderRequest) -> DetailedOrderResult:
+    def order(self, request: ArchiveOrderRequest) -> Order:
         '''
             Order from the Arlula imagery archive
         '''
@@ -506,4 +649,23 @@ class ArchiveAPI:
         if response.status_code != 200:
             raise ArlulaAPIException(response)
         else:
-            return DetailedOrderResult(json.loads(response.text))
+            return Order(json.loads(response.text))
+        
+    def batch_order(self, request: ArchiveBatchOrderRequest) -> Order:
+        '''
+            Order multiple scenes from the Arlula imagery archive
+        '''
+
+        url = self.url + "/order/batch"
+
+        response = requests.request(
+            "POST",
+            url,
+            data=json.dumps(request.dict()),
+            headers=self.session.header,
+        )
+
+        if response.status_code != 200:
+            raise ArlulaAPIException(response)
+        else:
+            return Order(json.loads(response.text))
