@@ -1,8 +1,8 @@
-import requests
+'''
+    Custom Exception Classes
+'''
 
-'''
-    Custom Exception Class
-'''
+import requests
 
 class ArlulaSessionError(Exception):
     def __init__(self, value):
@@ -14,6 +14,7 @@ class ArlulaSessionError(Exception):
 class ArlulaAPIException(Exception):
     def __init__(self, response: requests.Response):
         self.value = f"{response.status_code}: {response.text}"
+        self.response = response
 
     def __str__(self):
         return self.value
\ No newline at end of file
